use crate::config::{
    APP_BASE_ADDRESS, APP_SIZE_LIMIT, KERNEL_STACK_SIZE, MAX_APP_NUM, USER_STACK_SIZE,
};
use crate::log_information;
use crate::trap::context::TrapContext;
use core::arch::asm;

#[repr(align(4096))]
#[derive(Copy, Clone)]
struct KernelStack {
    data: [u8; KERNEL_STACK_SIZE],
}

#[repr(align(4096))]
#[derive(Copy, Clone)]
struct UserStack {
    data: [u8; USER_STACK_SIZE],
}

static KERNEL_STACK: [KernelStack; MAX_APP_NUM] = [KernelStack {
    data: [0; KERNEL_STACK_SIZE],
}; MAX_APP_NUM];

static USER_STACK: [UserStack; MAX_APP_NUM] = [UserStack {
    data: [0; USER_STACK_SIZE],
}; MAX_APP_NUM];

impl KernelStack {
    fn get_sp(&self) -> usize {
        self.data.as_ptr() as usize + KERNEL_STACK_SIZE
    }

    pub fn push_context(&self, trap_context: TrapContext) -> usize {
        let pointer = (self.get_sp() - size_of::<TrapContext>()) as *mut TrapContext;

        unsafe {
            *pointer = trap_context;
        }

        pointer as usize
    }
}

impl UserStack {
    fn get_sp(&self) -> usize {
        self.data.as_ptr() as usize + USER_STACK_SIZE
    }
}

pub fn get_app_num() -> usize {
    extern "C" {
        fn _num_app();
    }

    let app_num_ptr = _num_app as usize as *const usize;

    unsafe { app_num_ptr.read_volatile() }
}

pub fn load_apps() {
    extern "C" {
        fn _num_app();
    }

    let app_num_ptr = _num_app as usize as *const usize;
    let app_num = unsafe { app_num_ptr.read_volatile() };
    let app_start = unsafe { core::slice::from_raw_parts(app_num_ptr.add(1), app_num + 1) };

    for i in 0..app_num {
        let base_address = APP_BASE_ADDRESS + i * APP_SIZE_LIMIT;

        // 清除目标的内存区域
        for address in base_address..base_address + APP_SIZE_LIMIT {
            unsafe {
                (address as *mut u8).write_volatile(0);
            }
        }

        // 将程序从数据段复制过来
        let source = unsafe {
            core::slice::from_raw_parts(app_start[i] as *const u8, app_start[i + 1] - app_start[i])
        };
        let destination =
            unsafe { core::slice::from_raw_parts_mut(base_address as *mut u8, source.len()) };

        destination.copy_from_slice(source);
    }

    log_information!("Load {} applications in memory.", app_num);
    unsafe {
        asm!("fence.i");
    }
}

pub fn initialize_app_context(app_id: usize) -> usize {
    KERNEL_STACK[app_id].push_context(TrapContext::init_application_context(
        APP_BASE_ADDRESS + app_id * APP_SIZE_LIMIT,
        USER_STACK[app_id].get_sp(),
    ))
}
