use lazy_static::lazy_static;

use crate::config::MAX_APP_NUM;
use crate::loader::{get_app_num, initialize_app_context};
use crate::log_information;
use crate::sbi::shutdown;
use crate::sync::single_cell::SingleCell;
use crate::task::context::TaskContext;
use crate::task::switch::__switch;

mod context;
mod switch;

#[derive(Copy, Clone, PartialEq)]
pub enum TaskStatus {
    NotInitialized,
    Ready,
    Running,
    Exited,
}

#[derive(Copy, Clone)]
pub struct TaskControlBlock {
    pub status: TaskStatus,
    pub context: TaskContext,
}

pub struct TaskManagerInner {
    tasks: [TaskControlBlock; MAX_APP_NUM],
    current_task: usize,
}

pub struct TaskManager {
    app_number: usize,
    inner: SingleCell<TaskManagerInner>,
}

lazy_static! {
    static ref TASK_MANAGER: TaskManager = {
        let app_num = get_app_num();
        let mut tasks = [TaskControlBlock {
            context: TaskContext::new(),
            status: TaskStatus::NotInitialized,
        }; MAX_APP_NUM];

        for (i, task) in tasks.iter_mut().enumerate() {
            task.context = TaskContext::goto_restore(initialize_app_context(i));
            task.status = TaskStatus::Ready;
        }

        TaskManager {
            app_number: app_num,
            inner: {
                SingleCell::new(TaskManagerInner {
                    tasks,
                    current_task: 0,
                })
            },
        }
    };
}

pub fn run_first_task() {
    TASK_MANAGER.run_first_task();
}

pub fn suspend_current_and_run_next() {
    TASK_MANAGER.mark_current_suspended();
    TASK_MANAGER.run_next_task();
}

pub fn exit_current_and_run_next() {
    TASK_MANAGER.mark_current_exited();
    TASK_MANAGER.run_next_task();
}

impl TaskManager {
    /// status from running to ready
    fn mark_current_suspended(&self) {
        let mut inner = self.inner.exclusive_borrow();
        let current = inner.current_task;
        inner.tasks[current].status = TaskStatus::Ready;
    }

    /// status from running to exited
    fn mark_current_exited(&self) {
        let mut inner = self.inner.exclusive_borrow();
        let current = inner.current_task;
        inner.tasks[current].status = TaskStatus::Exited;
    }

    fn find_next_task(&self) -> Option<usize> {
        let inner = self.inner.exclusive_borrow();
        (inner.current_task + 1..inner.current_task + self.app_number + 1)
            .map(|id| id % self.app_number)
            .find(|id| inner.tasks[*id].status == TaskStatus::Ready)
    }

    fn run_next_task(&self) {
        if let Some(next) = self.find_next_task() {
            let mut inner = self.inner.exclusive_borrow();
            let current = inner.current_task;

            inner.tasks[next].status = TaskStatus::Running;
            inner.current_task = next;

            let current_task_context = &mut inner.tasks[current].context as *mut TaskContext;
            let next_task_context = &inner.tasks[next].context as *const TaskContext;
            drop(inner);

            unsafe {
                __switch(current_task_context, next_task_context);
            }
        } else {
            log_information!("All application completed!");
            shutdown(false);
        }
    }

    fn run_first_task(&self) {
        let mut inner = self.inner.exclusive_borrow();
        let task0 = &mut inner.tasks[0];
        task0.status = TaskStatus::Running;
        let next_task_context = &task0.context as *const TaskContext;
        drop(inner);

        let mut unused = TaskContext::new();
        unsafe {
            __switch(&mut unused as *mut TaskContext, next_task_context);
        }
    }
}
