use crate::log_warning;
use crate::syscall::syscall;
use crate::task::{exit_current_and_run_next, suspend_current_and_run_next};
use crate::timer::set_next_trigger;
use crate::trap::context::TrapContext;
use core::arch::global_asm;
use riscv::register::{
    mtvec::TrapMode,
    scause::{self, Exception, Interrupt, Trap},
    sie, stval, stvec,
};

pub mod context;

global_asm!(include_str!("trap.asm"));

pub fn init() {
    extern "C" {
        fn __alltraps();
    }
    unsafe {
        // 设置中断向量
        stvec::write(__alltraps as usize, TrapMode::Direct);

        // 启动时钟中断
        sie::set_stimer();
        set_next_trigger();
    }
}

#[no_mangle]
pub fn trap_handler(context: &mut TrapContext) -> &mut TrapContext {
    let scause = scause::read();
    let stval = stval::read();

    match scause.cause() {
        Trap::Exception(Exception::UserEnvCall) => {
            context.sepc += 4;
            context.x[10] = syscall(context.x[17], [context.x[10], context.x[11], context.x[12]]);
        }
        Trap::Exception(Exception::StoreFault) | Trap::Exception(Exception::StorePageFault) => {
            log_warning!("PageFault in application, kernel will kill it.");
            exit_current_and_run_next();
        }
        Trap::Exception(Exception::IllegalInstruction) => {
            log_warning!("Illegal instruction in application, kernel will kill it.");
            exit_current_and_run_next();
        }
        Trap::Exception(Exception::LoadFault) => {
            log_warning!("Illegal memory access in application, kernel will kill it.");
            exit_current_and_run_next();
        }
        Trap::Interrupt(Interrupt::SupervisorTimer) => {
            set_next_trigger();
            suspend_current_and_run_next();
        }
        _ => {
            panic!(
                "[Kernel] Unsupported trap: {:?}, stval = {:#x}!",
                scause.cause(),
                stval
            );
        }
    }

    context
}
